import pyadchpp as adchpp
import os, sys, traceback
import threading

hubThread = None

core = None
CfgDir = None
scriptDir = None
ClientMgr = adchpp.ClientManager
PluginMgr = adchpp.PluginManager
SocketMgr = adchpp.SocketManager
LogMgr = adchpp.SocketManager

loadedPrimaryModules = []

def initHub(configDirectory):
    global core, CfgDir, scriptDir, ClientMgr, PluginMgr, SocketMgr, LogMgr
    path = os.getcwd()
    CfgDir = configDirectory
    core = adchpp.Core.create(CfgDir+"/")
    scriptDir = 'PyScripts'

    ClientMgr = core.getClientManager()
    PluginMgr = core.getPluginManager()
    SocketMgr = core.getSocketManager()
    LogMgr = core.getLogManager()
    return loadHubXML(os.path.join(CfgDir,'adchpp.xml'))

def loadHubXML(fileName):
    try:
        xml = adchpp.SimpleXML()
        xml.fromXML( open(fileName, "r").read() )
        xml.resetCurrentChild()
        xml.stepIn()

        while xml.findChild(""):
            if xml.getChildName() == "Settings":
                xml.stepIn()
                while xml.findChild(""):
                    print("Processing %s" % xml.getChildName())
                    if xml.getChildName() == "HubName":
                        ClientMgr.getEntity(adchpp.AdcCommand.HUB_SID).setField("NI", xml.getChildData())
                    elif xml.getChildName() == "Description":
                        ClientMgr.getEntity(adchpp.AdcCommand.HUB_SID).setField("DE", xml.getChildData())
                    elif xml.getChildName() == "Log":
                        LogMgr.enabled = int(xml.getChildData())
                    elif xml.getChildName() == "LogFile":
                        LogMgr.logFile = xml.getChildData()
                    elif xml.getChildName() == "MaxCommandSize":
                        ClientMgr.setMaxCommandSize(int(xml.getChildData()))
                    elif xml.getChildName() == "BufferSize":
                        SocketMgr.setBufferSize(int(xml.getChildData()))
                    elif xml.getChildName() == "MaxBufferSize":
                        SocketMgr.setMaxBufferSize(int(xml.getChildData()))
                    elif xml.getChildName() == "OverflowTimeout":
                        SocketMgr.setOverflowTimeout(int(xml.getChildData()))
                    elif xml.getChildName() == "LogTimeout":
                        ClientMgr.setLogTimeout(int(xml.getChildData()))

                xml.stepOut()
            elif xml.getChildName() == "Servers":
                xml.stepIn()

                servers = adchpp.TServerInfoList()

                while xml.findChild("Server"):
                    server = adchpp.ServerInfo.create()
                    server.port = str(xml.getChildAttrib("Port"))
                    server.ip = "0.0.0.0"
                    #server.ip = adchpp.Util.getLocalIp()

                    if xml.getBoolChildAttrib("TLS"):
                        server.TLSParams.cert = os.path.abspath(os.path.join(CfgDir,xml.getChildAttrib("Certificate")))
                        server.TLSParams.pkey = os.path.abspath(os.path.join(CfgDir,xml.getChildAttrib("PrivateKey")))
                        server.TLSParams.trustedPath = os.path.abspath(os.path.join(CfgDir,xml.getChildAttrib("TrustedPath")))
                        server.TLSParams.dh = os.path.abspath(os.path.join(CfgDir,xml.getChildAttrib("DHParams")))

                        if not server.secure():
                            raise Exception("Error listening on port %s, server not secure\n" % server.port)
                        else:
                            servers.push_back(server)
                    else:
                        servers.push_back(server)

                SocketMgr.setServers(servers)

                xml.stepOut()
            elif xml.getChildName() == "Plugins":
                PluginMgr.setPluginPath(xml.getChildAttrib("Path"))
                xml.stepIn()
                plugins = []

                while xml.findChild("Plugin"):
                    plugins.append(xml.getChildData())

                PluginMgr.setPluginList(plugins)
                xml.stepOut()

        xml.stepOut()
        return True
    except Exception as e:
        print("Unable to load adchpp.xml. (%s)\n" % e)
    return False

def loadPrimaryModules():
    global scriptDir, loadedPrimaryModules
    try:
        scriptsToLoad = []
        
        xml = adchpp.SimpleXML()
        xml.fromXML( open(os.path.join(CfgDir, 'PyModules.xml'), "r").read())
        xml.resetCurrentChild()
        if xml.getChildName() == "PyModules":
            scriptDir = xml.getChildAttrib("Path")
        xml.stepIn()
        while xml.findChild(""):
            if xml.getChildName() == "Module":
                scriptsToLoad.append(xml.getChildData())

        if not os.path.exists(scriptDir):
            os.mkdir(scriptDir)
        if not scriptDir in sys.path:
            sys.path.insert(0, scriptDir)

        for script in scriptsToLoad:
            name, ext = os.path.splitext(script)
            if ext == '.py' or ext == '.pyc':
                try:
                    p = __import__(name, None, None, [''])
                    if hasattr(p, 'initModule'):
                        impfunc = getattr(p, 'initModule')
                        impfunc()
                        loadedPrimaryModules.append(p)
                        print('Loaded module: ' + p.__name__)
                    else:
                        print("Module " + p.__name__ + " has no 'initModule' function, not loaded")
                except Exception as err:
                    traceback.print_exc(file=sys.stdout)
    except Exception as err:
        traceback.print_exc(file=sys.stdout)

def runHub():
    core.run()
    #Hub shutting down here
    import UserManager, access_AccessHandle
    UserManager.saveUsers()
    access_AccessHandle.saltPlugData = None

def shutdownHub():
    global core, loadedPrimaryModules
    core.shutdown()
    del(core)

#not going to bother with this right now
#def restartHub():
#    global core, loadedPrimaryModules
#    core.shutdown()
#    del(core)
#    os.startfile(os.path.join(os.getcwd(),'adchpp.py') + ' -c ' + CfgDir)
#    sys.exit()

#reload doesn't actually work because of crappy python
#def reloadPrimaryModules(ent, cmd, params, public):
#    for mod in loadedPrimaryModules:
#        reload(mod)

#    while len(loadedPrimaryModules):
#        mod = loadedPrimaryModules.pop(0)
#        del(mod)

#    loadPrimaryModules()
#    import autil
#    autil.hub_reply(ent, "Modules reload action completed.")

def startHub():
    global hubThread
    try:
        loadPrimaryModules()

        hubThread = threading.Thread(target=runHub, args=())
        hubThread.start()
    except Exception as err:
        traceback.print_exc(file=sys.stdout)
