import autil
import Hub
import CommandManager
import json
import os
import threading, time

settings = {}
settingsLock = threading.RLock()

settingsFile = 'settings.txt'
fileLock = threading.RLock()

initiated = False

class Setting(dict):
    def __init__(self, key = None, alias = [], change = [], helpTxt = "", value = None, validate = None):        
        self['key'] = key
        self['alias'] = alias
        if type(alias) is list:
            self['alias'] = alias
        elif type(alias) is str:
            self['alias'] = [alias]
        else:
            self['alias'] = []
        if type(change) is list:
            self['change'] = change
        elif callable(change):
            self['change'] = [change]
        else:
            self['change'] = []
        #self['change'] = change
        self['helpTxt'] = helpTxt
        self['value'] = value
        self['validate'] = validate

    def release(self):
        self.change = []
        self.validate = None

    key = property(lambda self: self['key'], (lambda self,x: self.__setitem__('key', x)))
    alias = property(lambda self: self['alias'], (lambda self,x: self.__setitem__('alias', x)))
    change = property(lambda self: self['change'], (lambda self,x: self.__setitem__('change', x)))
    helpTxt = property(lambda self: self['helpTxt'], (lambda self,x: self.__setitem__('helpTxt', x)))
    value = property(lambda self: self['value'], (lambda self,x: self.__setitem__('value', x)))
    validate = property(lambda self: self['validate'], (lambda self,x: self.__setitem__('validate', x)))

def initModule():
    global settings, initiated
    fileLock.acquire()
    settingsLock.acquire()
    readSettings = autil.parseJsonFile(os.path.join(Hub.CfgDir, settingsFile))
    oldSettings = settings.copy()
    if readSettings:
        for k in readSettings:
            settings[k] = Setting(key = k, value = readSettings[k])
            settings[k]['scriptAdded'] = False

    if oldSettings:
        settings.update(oldSettings)

    fileLock.release()
    settingsLock.release()
    CommandManager.add(CommandManager.Command(key="cfg",alias=['config','var','changevar','setvar','setcfg','setconfig'],category="HubManagement",command=cfgCommand,minLevel=10,helpTxt='name value - change hub configuration, use "help cfg" to list all variables',helpTxtLong=listSettings))

    initiated = True

def listSettings():
    settingsStr = "List of all settings variables:\n"
    setts = settings.items()
    setts.sort()
    numUnset = 0
    for k in setts:
        #print(k[0])
        key = k[0]
        value = k[1]
        if value['scriptAdded']:
            settingsStr += key + " - current value: " + str(value.value)
            if value.helpTxt and len(value.helpTxt) > 0:
                settingsStr += " - " + value.helpTxt
            if value.alias and len(value.alias) > 0:
                settingsStr += " (aliases:  "
                for alias in value.alias:
                    settingsStr += alias + ", "
                settingsStr = settingsStr[:len(settingsStr)-2] + ")"
            settingsStr += "\n"
        else:
            numUnset += 1

    if numUnset > 0:
        settingsStr += "\n\nThe below settings have only been read from the settings file, they may not have any function:\n" 
        for k in setts:
            key = k[0]
            value = k[1]
            if not value['scriptAdded']:
                settingsStr += key + " - current value: " + str(value.value)
                if value.helpTxt and len(value.helpTxt) > 0:
                    settingsStr += " - " + value.helpTxt
                if value.alias and len(value.alias) > 0:
                    settingsStr += " (aliases: "
                    for alias in value.alias:
                        settingsStr += alias + ", "
                    settingsStr = settingsStr[:len(settingsStr)-2] + ")"
                settingsStr += "\n"
    return(settingsStr)

def cfgCommand(ent, cmd, params, public):
    if len(params) == 0:
        autil.hub_reply(ent, "You need to supply a variable name")
        return False

    key = None
    params = CommandManager.strSplitFirstWhite(params[0])
    
    for k in settings:
        if params[0] == k or (params[0] in settings[k].alias):
            key = k
            break
        
    if not key:
        autil.hub_reply(ent, "The name " + params[0] + ''' doesn't correspond to any setting variable, use "''' + str(settings['cmd_chr'].value) + '''help cfg " to list all variables''')
        return False

    oldValue = settings[key].value
    oldType = type(oldValue)

    params.pop(0)

    newValue = None
    
    settingsLock.acquire()
    
    if len(params) == 0 or len(params[0]) == 0:
        #no value; make up a default one
        if 'default' in settings[key]:
            newValue = settings[key]['default']
        else:
            newValue = ""
    else:
        newValue = params[0]

    try:
        newValue = oldType(newValue)
    except:
        autil.hub_reply(ent, "The value \"" + newValue + "\" could not be cast properly, enter a value of type " + str(oldType) + " for setting: " + key)
        settingsLock.release()
        return False

    if newValue == oldValue:
        autil.hub_reply(ent, "The value is the same as before, no change done")
        settingsLock.release()
        return False

    if settings[key].validate:
        err = settings[key].validate(newValue)
        if err:
            autil.hub_reply(ent, "The new value \"" + newValue + "\" is invalid, no change done (" + err + ")")
            settingsLock.release()
            return False

    settings[key].value = newValue
    if settings[key].change:
        for func in settings[key].change:
            func(oldValue, newValue)

    #time.sleep(0.1)
    settingsLock.release()
    save()
    
    autil.hub_reply(ent, "The new value \"" + str(newValue) + "\" has been successfully set for \"" + key + "\"")

    message = ent.getField('NI') + ' has changed "' + key + '" from "' + str(oldValue) + '" to "' + str(newValue) + '"'
    autil.log(__name__, message)
    #cm:sendToAll(autil.info(message):getBuffer())
        
def save():
    if initiated:
        toSave = {}
        for k in settings:
                toSave[k] = settings[k].value
        fileLock.acquire()
        autil.saveJsonFile(toSave, os.path.join(Hub.CfgDir, settingsFile))
        fileLock.release()

def add(setting):
    change = False
    new = False
    setting['default'] = setting.value

    settingsLock.acquire()    
    if setting.key != None and type(setting.key) is str and len(setting.key) > 0:
        if setting.key in settings:
            currValue = settings[setting.key].value
            try:
                currValue = type(setting.value)(currValue)
            except Exception as err:
                autil.log(__name__,"Setting \"" + setting.key + "\" could not be type cast when added, reverted to default value: " + str(options.value) + "  ~  (" + str(err) + ")")
                currValue = setting.value
                
            change = currValue != setting.value
            setting.value = currValue
        else:
            new = True
    else:
        autil.log(__name__, "Setting could not be added because it had no key")

    settings[setting.key] = setting
    settings[setting.key]['scriptAdded'] = True

    if setting.change:
        for func in setting.change:
            func(setting['default'], setting.value)

    if new:
        save()

    settingsLock.release()
