import CommandManager
import SettingManager
import pyadchpp as adchpp
import Hub
import autil
import random
from ScriptManager import ScriptBase
import threading, os, sys

ballFile = os.path.join('/scriptData/','8ball.txt')
fileLock = threading.RLock()

class Script(ScriptBase):

    default_answers=["As I see it, yes","It is certain","It is decidedly so","Most likely","Outlook good","Signs point to yes", "Without a doubt", "Yes", "Yes - definitely","You may rely on it", "Reply hazy, try again","Ask again later", "Better not tell you now","Cannot predict now","Concentrate and ask again","Don't count on it","My reply is no","My sources say no","Outlook not so good","Very doubtful"]
    answers = []
    answersLock = threading.RLock()

    def __init__(self):
        random.seed()

        fileLock.acquire()
        self.answersLock.acquire()
        self.answers = autil.parseJsonFile(os.path.join(Hub.CfgDir, ballFile))

        if not self.answers:
            self.answers = list(self.default_answers)
            self.saveAnswers()

        fileLock.release()
        self.answersLock.release()

        self.ballCmd = CommandManager.Command(key='8ball', alias=['magic8ball'],category="ScriptCommands/8BallCommands",command=self.getAnswer, helpTxt="[string] - Queries the great and all knowing Magic 8-Ball", user_command=
                                                  CommandManager.UserCommand(contexts=CommandManager.context_hub+CommandManager.context_userlist, path="Utility/Magic 8-Ball", hasPublic=True, command="8ball " + CommandManager.createLine("Query")))
        self.listCmd = CommandManager.Command(key='8list',category="ScriptCommands/8BallCommands",command=self.listAnswers, helpTxt="Returns a list of all 8ball available answers")
        self.defaultCmd = CommandManager.Command(key='8default',category="ScriptCommands/8BallCommands",command=self.defaultAnswers, helpTxt="Retruns all of the 8ball answers to their standard set", minLevel=SettingManager.settings['level_op'])
        self.appendCmd = CommandManager.Command(key='8append',category="ScriptCommands/8BallCommands",command=self.appendAnswer, helpTxt="string - Appends the string to the list of answers", minLevel=SettingManager.settings['level_op'])
        self.removeCmd = CommandManager.Command(key='8remove',category="ScriptCommands/8BallCommands",command=self.removeAnswer, helpTxt="int - Removes the answer at index [int] from the list of answers", minLevel=SettingManager.settings['level_op'])
        CommandManager.add(self.ballCmd)
        CommandManager.add(self.listCmd)
        CommandManager.add(self.defaultCmd)
        CommandManager.add(self.appendCmd)
        CommandManager.add(self.removeCmd)

    def reload(self):
        self.ballCmd.command = self.getAnswer
        self.listCmd.command = self.listAnswers
        self.defaultCmd.command = self.defaultAnswers
        self.appendCmd.command = self.appendAnswer
        self.removeCmd.command = self.removeAnswer

    def unload(self):
        self.ballCmd.release()
        self.listCmd.release()
        self.defaultCmd.release()
        self.appendCmd.release()
        self.removeCmd.release()

    def saveAnswers(self):
        fileLock.acquire()
        autil.saveJsonFile(self.answers, Hub.CfgDir + ballFile)
        fileLock.release()

    def listAnswers(self, ent, cmd, params, public):
        strAnswers = "8Ball answers:\n"
        for i in range(len(self.answers)):
            strAnswers += "[" + str(i) + "]: " + str(self.answers[i]) + "\n"

        autil.hub_reply(ent, strAnswers)

    def defaultAnswers(self, ent, cmd, params, public):
        self.answersLock.acquire()
        self.answers = list(self.default_answers)
        self.saveAnswers()
        self.answersLock.release()
        autil.hub_reply(ent, "Answer successfully set back to default 8ball answers.")

    def removeAnswer(self, ent, cmd, params, public):
        if not params:
            autil.hub_reply(ent, "You need to supply an index to remove.")

        num = autil.toInt(params[0])

        if num == None or num not in range(len(self.answers)):
            autil.hub_reply(ent, "You need to supply an integer index of an existing answer.")

        self.answersLock.acquire()
        try:
            
            del(self.answers[num])
            autil.hub_reply(ent, "Answer successfully removed.")
            self.saveAnswers()
        except Exception as err:
            autil.hub_reply(ent, "Answer could not be removed. (" + str(err) + ")")
        self.answersLock.release()

    def appendAnswer(self, ent, cmd, params, public):
        if not params:
            autil.hub_reply(ent, "You need to supply an answer to append.")

        self.answersLock.acquire()
        try:
            
            self.answers.append(str(params[0]))
            autil.hub_reply(ent, "Answer successfully appended.")
            self.saveAnswers()
        except Exception as err:
            autil.hub_reply(ent, "Answer could not be appended. (" + str(err) + ")")
        self.answersLock.release()

    def getAnswer(self, ent, cmd, params, public):
        returnMsg = ""
        if params:
            returnMsg += '"' + params[0] + '": '
        
        if self.answers:
            returnMsg += str(self.answers[random.randint(0,len(self.answers)-1)])

        if public:
            autil.ent_reply(ent, "[Magic 8-Ball] " + returnMsg);
        else:
            autil.hub_reply(ent, returnMsg);