import CommandManager
import pyadchpp as adchpp
import Hub
import UserManager
import SettingManager
from ScriptManager import ScriptBase
import autil
import threading
import time, os, json

class Script(ScriptBase):
    history = []
    historyDir = os.path.abspath(os.path.join(Hub.CfgDir , "history"))
    historyFmt = "history_%Y_%m_%d.txt"
    historyLock = threading.Lock()
    toSaveHistory = []
    saveHistoryLock = threading.Lock()

    topicHistory = []
    topicHistoryFile = os.path.join(Hub.CfgDir , "topicHistory.txt")
    topicHistoryLock = threading.Lock()
    toSaveTopicHistory = []
    saveTopicHistoryLock = threading.Lock()

    histCommand = None
    topicHistCommand = None

    def __init__(self):
        self.hidden_bot = None
        self.history_1 = None

        self.changeDescFunc = self.changeDesc;
        self.changeHistorySettingFunc = self.changeHistorySetting
        self.methodChangeFunc = self.changeHistorySetting
        self.changeTopicHistorySettingFunc = self.changeHistorySetting
        
        self.history_enable = SettingManager.Setting(key="history_enable",change=self.changeHistorySettingFunc,helpTxt="keep history of chat, 0 = disabled, 1 = enabled",value=1)
        self.history_archive = SettingManager.Setting(key="history_archive",helpTxt="enable archiving of history using history_prefix format, 0 = disabled, 1 = enabled",value=0)
        self.history_default = SettingManager.Setting(key="history_default",helpTxt="number of messages to display in history if the user doesn't select anything else",value=50)
        self.history_max = SettingManager.Setting(key="history_max",helpTxt="number of messages to keep for history",value=500)
        self.history_method = SettingManager.Setting(key="history_method",change=self.methodChangeFunc,helpTxt="strategy used by the history script to record messages, restart the hub after the change, 1 = use a hidden bot, 0 = direct ADCH++ interface",value=1)
        self.history_prefix = SettingManager.Setting(key="history_prefix",helpTxt="prefix to put before each message in history",value="[%Y-%m-%d %H:%M:%S] ")
        self.history_address = SettingManager.Setting(key="history_address",helpTxt="address to direct users to for a history archives",value="")

        self.cmdClearCache = CommandManager.Command(key="clearHistoryCache",command=self.clearCacheCMD,category="ScriptCommands", helpTxt="- clears all the messages currently stored in the cache", minLevel=SettingManager.settings['level_op'].value, user_command=
                                                  CommandManager.UserCommand(contexts=CommandManager.context_hub+CommandManager.context_userlist, path="Clear Chat History", command="clearHistoryCache"))
        CommandManager.add(self.cmdClearCache)
        
        SettingManager.add(self.history_enable)
        SettingManager.add(self.history_archive)
        SettingManager.add(self.history_default)
        SettingManager.add(self.history_max)
        SettingManager.add(self.history_method)
        SettingManager.add(self.history_prefix)
        SettingManager.add(self.history_address)

        self.history_topic_enable = SettingManager.Setting(key="history_topic_enable",change=self.changeTopicHistorySettingFunc,helpTxt="keep history of past topics and who set them, 0 = disabled, 1 = enabled",value=1)
        self.history_topic_archive = SettingManager.Setting(key="history_topic_archive",helpTxt="enable archiving of topic history, 0 = disabled, 1 = enabled",value=0)
        self.history_topic_default = SettingManager.Setting(key="history_topic_default",helpTxt="number of messages to display in history if the user doesn't select anything else",value=10)
        self.history_topic_max = SettingManager.Setting(key="history_topic_max",helpTxt="number of messages to keep for history",value=50)
        self.history_topic_prefix = SettingManager.Setting(key="history_topic_prefix",helpTxt="prefix to put before each topic in history",value="[%Y-%m-%d %H:%M:%S] ")
        self.history_topic_address = SettingManager.Setting(key="history_topic_address",helpTxt="address to direct users to for a topic history archives",value="")
        SettingManager.add(self.history_topic_enable)
        SettingManager.add(self.history_topic_archive)
        SettingManager.add(self.history_topic_default)
        SettingManager.add(self.history_topic_max)
        SettingManager.add(self.history_topic_prefix)
        SettingManager.add(self.history_topic_address)

    def createBotListener(self):
        self.hidden_bot = Hub.ClientMgr.createBot(lambda bot, buffer: self.parse(None, adchpp.AdcCommand(buffer), True))
        self.hidden_bot.setField('ID', self.hidden_bot.getCID().toBase32())
        self.hidden_bot.setField('NI', __name__ + '-hidden_bot')
        self.hidden_bot.setField('DE', "Hidden bot used by the " + __name__ + " script")
        self.hidden_bot.setFlag(adchpp.Entity.FLAG_HIDDEN)
        Hub.ClientMgr.regBot(self.hidden_bot)

    def methodChange(self, oldValue, newValue):
        if oldValue == 0:
            self.history_1 = None
        elif self.hidden_bot:
            self.hidden_bot.disconnect(adchpp.Util.REASON_PLUGIN)

        if newValue == 0:
            self.history_1 = autil.handleCommand(self.parse, adchpp.AdcCommand.CMD_MSG)
        else:
            self.createBotListener()

    def changeHistorySetting(self, oldValue, newValue):
        if not self.histCommand and newValue:
            self.histCommand = CommandManager.Command(key="history",command=self.histCMD,category="ScriptCommands", helpTxt="[lines] - display main chat messages logged by the hub (no lines=default / since last logoff)", user_command=
                                                  CommandManager.UserCommand(contexts=CommandManager.context_hub+CommandManager.context_userlist, path="Chat History", command="history " + CommandManager.createLine("Number of msg's to display (empty=default / since last logoff)")))
            CommandManager.add(self.histCommand)

            if self.history_method.value == 0:
                self.history_1 = autil.handleCommand(self.parse, adchpp.AdcCommand.CMD_MSG)
            else:
                self.createBotListener()

        elif self.histCommand and not newValue:
            self.histCommand.release()
            self.histCommand = None

            if self.history_method.value == 0:
                self.history_1 = None
            elif self.hidden_bot:
                self.hidden_bot.disconnect(adchpp.Util.REASON_PLUGIN)

    def changeTopicHistorySetting(self, oldValue, newValue):
        if not self.topicHistCommand and newValue:
            self.topicHistCommand = CommandManager.Command(key="topicHistory",alias=['tHistory'],command=self.topicHistCMD,category="ScriptCommands", helpTxt="[number] - display topics logged by the hub (no number=default)", user_command=
                                                    CommandManager.UserCommand(contexts=CommandManager.context_hub+CommandManager.context_userlist, path="Topic History", command="topicHistory " + CommandManager.createLine("Number of topics's to display (empty=default)")))
            CommandManager.add(self.topicHistCommand)

            SettingManager.settings['topic'].change.append(self.changeDesc)

        elif self.topicHistCommand and not newValue:
            self.topicHistCommand.release()
            self.topicHistCommand = None
            SettingManager.settings['topic'].change.remove(self.changeDesc)
            

    def changeDesc(self, oldValue, newValue):
        topic = {'time':0,'topic':newValue,'nick':"Unknown"}
        if hasattr(autil.threadData, 'origUser'):
            topic['nick'] = autil.threadData.origUser.nick
        topic['time'] = time.time()
        self.topicHistory.append(topic)

        if self.history_topic_archive.value == 1:
            self.saveTopicHistoryLock.acquire()
            self.toSaveTopicHistory.append(topic)
            self.saveTopicHistoryLock.release()
            self.saveTopicThread = threading.Thread(target=self.saveTopicHistory)
            self.saveTopicThread.run()
        
        if len(self.topicHistory) > self.history_topic_max.value:
            self.topicHistory = self.topicHistory[len(self.topicHistory)-self.history_topic_max.value:len(self.topicHistory)]

    def topicHistCMD(self, ent, cmd, params, public):
        items = None
        if len(params) > 0 and len(params[0]) > 0:
            try:
                items = int(params[0])
            except:
                autil.hub_reply(ent, "Parameter '" + str(params[0]) + "' could not be cast as int")
                return

        if not items:
            items = self.history_topic_default.value

        if items > self.history_topic_max.value:
            items = self.history_topic_max.value

        startPos = 0
        
        if items > len(self.topicHistory):
            startPos = len(self.topicHistory)
        else:
            startPos = items

        msg = "Displaying the last " + str(startPos) + " topics"

        for hist in self.topicHistory[len(self.topicHistory)-startPos:len(self.topicHistory)]:
            msg += "\r\n" + hist['topic'] + " set by: " + hist['nick'] + " on " + time.strftime(self.history_topic_prefix.value, time.localtime(hist['time']))

        if self.history_topic_address.value:
            msg += "\n\nTopic history archives are available at: " + self.history_topic_address.value

        autil.hub_reply(ent, msg)

    def saveTopicHistory(self):
        if self.history_topic_archive.value == 1 and self.toSaveTopicHistory:
            self.saveTopicHistoryLock.acquire()
            toSaveTopicHistory = self.toSaveTopicHistory
            self.toSaveTopicHistory = []
            self.saveTopicHistoryLock.release()

            if not os.path.exists(self.historyDir):
                os.makedirs(self.historyDir)
            append = False
            try:
                if os.path.exists(self.topicHistoryFile):
                    append = True
                    histfile = open(self.topicHistoryFile, 'r+')
                else:
                    histfile = open(self.topicHistoryFile, 'w')
            
                if append:
                    histfile.seek(-1,os.SEEK_END)
                    if histfile.read() == "]":
                        histfile.seek(-1,os.SEEK_END)
                        histfile.write(", ")
                        histfile.write(json.dumps(toSaveTopicHistory)[1:])
                    else:
                        histfile.close()
                        os.rename(self.topicHistoryFile, self.topicHistoryFile+"_corrupt")
                        histfile = open(self.topicHistoryFile, 'w')
                        json.dump(toSaveTopicHistory, histfile)
                else:
                    json.dump(toSaveTopicHistory, histfile)
            except Exception as err:
                autil.log(__name__, self.topicHistoryFile + ' could not be saved (' + str(err) + ')')

    def clearCacheCMD(self, ent, cmd, params, public):
        sdf

    def saveHistory(self):
        if self.history_archive.value == 1 and self.toSaveHistory:
            self.saveHistoryLock.acquire()
            toSaveHistory = self.toSaveHistory
            self.toSaveHistory = []
            self.saveHistoryLock.release()

            if not os.path.exists(self.historyDir):
                os.makedirs(self.historyDir)
            histfilePath = os.path.abspath( os.path.join(self.historyDir , str(time.strftime(self.historyFmt,time.localtime(time.time())))) )
            append = False
            try:
                if os.path.exists(histfilePath):
                    append = True
                    histfile = open(histfilePath, 'r+')
                else:
                    histfile = open(histfilePath, 'w')
            
                if append:
                    histfile.seek(-1,os.SEEK_END)
                    if histfile.read() == "]":
                        histfile.seek(-1,os.SEEK_END)
                        histfile.write(", ")
                        histfile.write(json.dumps(toSaveHistory)[1:])
                    else:
                        histfile.close()
                        os.rename(histfilePath, histfilePath+"_corrupt")
                        histfile = open(histfilePath, 'w')
                        json.dump(toSaveHistory, histfile)
                else:
                    json.dump(toSaveHistory, histfile)
            except Exception as err:
                autil.log(__name__, histfilePath + ' could not be saved (' + str(err) + ')')

    def unload(self):
        if self.history_method.value == 0:
            self.history_1 = None
        elif self.hidden_bot:
            self.hidden_bot.disconnect(adchpp.Util.REASON_PLUGIN)
            self.hidden_bot = None

        self.histCommand.release()
        self.cmdClearCache.release()

        self.history_enable.change.remove(self.changeHistorySetting)
        self.history_method.change.remove(self.methodChange)
        self.history_topic_enable.change.remove(self.changeTopicHistorySetting)

        if self.topicHistCommand:
            self.topicHistCommand.release()
            self.topicHistCommand = None
            SettingManager.settings['topic'].change.remove(self.changeDesc)

    def reload(self):
        if self.history_method.value == 0:
            self.history_1 = None
            self.history_1 = autil.handleCommand(self.parse, adchpp.AdcCommand.CMD_MSG)
        else:
            if self.hidden_bot:
                self.hidden_bot.disconnect(adchpp.Util.REASON_PLUGIN)
                self.hidden_bot = None
            self.createBotListener()

        self.histCommand.command = self.histCMD
        self.cmdClearCache.command = self.clearCacheCMD

        self.history_enable.change.remove(self.changeHistorySettingFunc)
        self.history_method.change.remove(self.methodChangeFunc)
        self.history_topic_enable.change.remove(self.changeTopicHistorySettingFunc)

        self.changeHistorySettingFunc = self.changeHistorySetting
        self.methodChangeFunc = self.changeHistorySetting
        self.changeTopicHistorySettingFunc = self.changeHistorySetting

        self.history_enable.change.append(self.changeHistorySettingFunc)
        self.history_method.change.append(self.methodChangeFunc)
        self.history_topic_enable.change.append(self.changeTopicHistorySettingFunc)

        if self.topicHistCommand:
            self.topicHistCommand.command = self.topicHistCMD
            SettingManager.settings['topic'].change.remove(self.changeDescFunc)
            SettingManager.settings['topic'].change.append(self.changeDesc)
        self.changeDescFunc = self.changeDesc

    def parse(self, ent, cmd, ok):
        if cmd.getCommand() != adchpp.AdcCommand.CMD_MSG or cmd.getType() != adchpp.AdcCommand.TYPE_BROADCAST:
            return True 

        if not ent:
            ent = Hub.ClientMgr.getEntity(cmd.getFrom())

        if not ent:
            return True
        
        nick = ent.getField("NI")

        if len(nick) < 1:
            return True

        isME, mepar = cmd.getParam("ME",0)

        hist = {'time':0,'message':"",'nick':""}
        hist['nick'] = nick
        hist['time'] = time.time()
        hist['me'] = isME
        hist['message'] = cmd.getParam(0)
        self.history.append(hist)

        if self.history_archive.value == 1:
            self.saveHistoryLock.acquire()
            self.toSaveHistory.append(hist)
            self.saveHistoryLock.release()
            self.saveThread = threading.Thread(target=self.saveHistory)
            self.saveThread.run()
        
        if len(self.history) > self.history_max.value:
            self.history = self.history[len(self.history)-self.history_max.value:len(self.history)]

    def get_items(self, ent):
        user = UserManager.findUser(ent.getField("NI"),ent.getCID().toBase32())
        if user:
            lasttime = user.lastofftime
            items = 0
            if lasttime:
                for hist in self.history:
                    if hist['time'] > lasttime:
                        break
                    else:
                        items += 1
            return len(self.history) - items

    def getHistoryLines(self, items):
        startPos = 0
        
        if items > len(self.history):
            startPos = len(self.history)
        else:
            startPos = items

        msg = "Displaying the last " + str(startPos) + " messages"

        for hist in self.history[len(self.history)-startPos:len(self.history)]:
            msg += "\r\n" + time.strftime(self.history_prefix.value, time.localtime(hist['time']))
            if 'me' in hist and hist['me']:
                msg +="* " + hist['nick'] + " "
            else:
                msg += "<" + hist['nick'] + "> "
            msg += hist['message']

        if self.history_address.value:
            msg += "\n\nHistory archives are available at: " + self.history_address.value

        return msg

    def histCMD(self, ent, cmd, params, public):
        items = None
        if len(params) > 0 and len(params[0]) > 0:
            try:
                items = int(params[0])
            except:
                autil.hub_reply(ent, "Parameter '" + str(params[0]) + "' could not be cast as int")
                return
        else:
            if UserManager.getLevel(ent) > 0:
                items = self.get_items(ent)

        if not items:
            items = self.history_default.value

        if items > self.history_max.value:
            items = self.history_max.value

        msg = self.getHistoryLines(items)

        autil.hub_reply(ent, msg)